#!/usr/bin/env python3
import socketserver as sock
import time
import threading
from Crypto.PublicKey import RSA
import math
import base64
import sys

from Crypto import Random
from Crypto.Math.Numbers import Integer
from Crypto.Math.Primality import (test_probable_prime, generate_probable_prime, COMPOSITE)
from Crypto.PublicKey.RSA import RsaKey

logging = False



FLAG = sys.argv[1].encode()
#FLAG = "flag{test_flag}".encode()

'''
Modified version of Crypto.PublicKey.RSA.generate

'''
def generate(bits, randfunc=None, e=65537):

    if bits < 1024:
        raise ValueError("RSA modulus length must be >= 1024")
    if e % 2 == 0 or e < 3:
        raise ValueError("RSA public exponent must be a positive, odd integer larger than 2.")

    if randfunc is None:
        randfunc = Random.get_random_bytes

    d = n = Integer(1)
    e = Integer(e)

    while n.size_in_bits() != bits and d < (1 << (bits // 2)):
        size_q = bits // 2
        size_p = bits - size_q

        min_p = min_q = (Integer(1) << (2 * size_q - 1)).sqrt()
        if size_q != size_p:
            min_p = (Integer(1) << (2 * size_p - 1)).sqrt()

        def filter_p(candidate):
            return candidate > min_p and (candidate - 1).gcd(e) == 1 and (candidate - 1).gcd(Integer(257)) == 1

        p = generate_probable_prime(exact_bits=size_p,
                                    randfunc=randfunc,
                                    prime_filter=filter_p)

        min_distance = Integer(1) << (bits // 2 - 100)

        def filter_q(candidate):
            return (candidate > min_q and
                    (candidate - 1).gcd(e) == 1 and (candidate - 1).gcd(Integer(257)) == 1 and
                    abs(candidate - p) > min_distance)

        q = generate_probable_prime(exact_bits=size_q,
                                    randfunc=randfunc,
                                    prime_filter=filter_q)

        n = p * q
        lcm = (p - 1).lcm(q - 1)
        d = e.inverse(lcm)

    if p > q:
        p, q = q, p

    u = p.inverse(q)

    return RsaKey(n=n, e=e, d=d, p=p, q=q, u=u)

def print_text(text):
    print(text, flush=True)

def init():
    keyPair = generate(2048)
    return keyPair

def asd(e, start):
    keyPair = RSA.construct((start.n, e))
    pubKey = keyPair.publickey()
    #print_text(pubKey.exportKey().decode('ascii'))
    return keyPair

def encrypt(e, n):
    keylength = math.ceil(n.bit_length() / 8)
    input_nr = int.from_bytes(FLAG, byteorder='big')
    crypted_nr = pow(input_nr, e, n)
    crypted_data = crypted_nr.to_bytes(keylength, byteorder='big')
    return base64.b64encode(crypted_data).decode()

class Service(sock.BaseRequestHandler):
    allow_reuse_address = True
    
    # Connection handler
    def handle(self):
        if logging:
            print_text("[+] Incoming connection")

        start = init()

        keyPair = asd(257, start)
        self.send("n = " + str(keyPair.publickey().n))
        self.send("e = " + str(keyPair.publickey().e))
        enc_text = encrypt(257, keyPair.publickey().n)
        self.send("Assignment: " + enc_text)
        self.send("")

        keyPair = asd(65537, start)
        self.send("n = " + str(keyPair.publickey().n))
        self.send("e = " + str(keyPair.publickey().e))
        enc_text = encrypt(65537, keyPair.publickey().n)
        self.send("Assignment: " + enc_text)

       
        
        self.send("")
        self.send("Gather at the cave. Br4in said they're ready for phase 2.")
        #response = self.receive()

    def send(self, text, newline=True):
        if logging:
            print_text("[>] "+ text)
        if newline: 
            text += "\n"
        self.request.sendall(text.encode())

    def receive(self):
        incoming = self.request.recv(4096).strip().decode('ASCII')
        if logging:
            print_text("[<] "+ incoming)
        return incoming
        
    
class ThreadService(sock.ThreadingMixIn, sock.TCPServer, sock.DatagramRequestHandler):
    pass
    

def main():
    host = sys.argv[2]
    port = int(sys.argv[3])
    #host = '127.0.0.1'
    #port = 1339
    
    s = Service
    server = ThreadService((host, port), s)

    server_thread = threading.Thread(target=server.serve_forever)
    server_thread.daemon = True
    server_thread.start()

    print_text("[Server started on: " + host + ":" + str(port) + "]")

    while(True): 
        time.sleep(1)


if (__name__=="__main__"): 
    main()

